#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <intrin.h>

//#define CACHESIZE (32*1024)
#define CACHESIZE (16*1024)

extern "C" {

typedef unsigned int     uint32_t;
typedef unsigned __int64 uint64_t;
typedef uint32_t (*benchfn_t)(void *arg);

void bench(const char *name, benchfn_t fn, void *arg)
{
	char buf[64];
	uint64_t before, after;
	uint32_t count;

	Sleep(1000);
	fn(arg);
	before = __rdtsc();
	count = fn(arg);
	after = __rdtsc();

	sprintf(buf, "%-20s %p", name, arg);
	printf("%-33s %6.3f\n", buf, ((double)(after - before)) / ((double)count));
}

#define BENCH(x, a)						\
	do {								\
		uint32_t __cdecl x(void *);				\
		bench(#x, x, a);				\
	} while(0)

int main(int argc, char **argv)
{
	char *p;
	char *q;

	SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
	SetThreadAffinityMask(GetCurrentThread(), 1);

	p = (char *)VirtualAlloc(NULL, CACHESIZE, MEM_COMMIT, PAGE_READWRITE);
	for (int i = 0; i < CACHESIZE; i+=4)
		*(DWORD *)(p+i) = (i * 3 + 4) % CACHESIZE;
	q = (char *)VirtualAlloc(NULL, CACHESIZE, MEM_COMMIT, PAGE_READWRITE);
	for (int i = 0; i < CACHESIZE; i+=8)
		*(__int64 *)(q+i) = (i * 3 + 8) % CACHESIZE;

	printf("algorithm                         cycles\n");
	printf("----------------------------------------\n");
	BENCH(vpgatherdd_4_tp, p);
	BENCH(vpgatherdd_8_tp, p);
	BENCH(vpgatherqq_2_tp, q);
	BENCH(vpgatherqq_4_tp, q);
	BENCH(vpgatherdd_8_half_tp, p);
	printf("----------------------------------------\n");
	BENCH(vpgatherdd_4_lt, p);
	BENCH(vpgatherdd_8_lt, p);
	BENCH(vpgatherqq_2_lt, q);
	BENCH(vpgatherqq_4_lt, q);
	BENCH(vpgatherdd_8_half_lt, p);
	printf("----------------------------------------\n");
	BENCH(insextdd_4_lt, p);
	BENCH(insextdd_8_lt, p);
	printf("----------------------------------------\n");

	return 0;
}

}